# Two modes:
#   (1) interactive: Everything runs in one thread. Long-running event handlers block the update.
#   (2) separated: non-main thread for all non-mainloop code.
"""
We match the Tk hierarchy of classes onto the ACM hierarchy as follows:
Each application (i.e. instance of TkBackend) has exactly one Tk root (i.e. one tkinter.Tk()).

This top widget takes care of everything.

After instantiation of the first instance of a TkBackend, the tkinter._default_root will be set.

Except in extreme circumstances, there should only ever be one instance of TkBackend.

Each GWindow is associated to a single Canvas. GWindows beyond the first will open a new Toplevel window.

## EXIT BEHAVIOR
(A) At program termination, Tk object is not destroyed, and "program" isn't done until Tk window is closed.
(B) At program termination, Tk object is destroyed and program is finished.
"""
# It is discouraged to instantiate multiple instances of Tk graphics
from campy.private.backends.backend_base import GraphicsBackendBase
from campy.private.backends.tk.menu import setup_menubar

import atexit
import logging
import tkinter as tk
import tkinter.font as tkfont
import tkinter.filedialog as tkfiledialog
import tkinter.messagebox as tkmessagebox
import tkinter.simpledialog as tksimpledialog
import threading
import sys

# TODO(sredmond): What magic is this?
try:
    from _tkinter import DONT_WAIT
except ImportError:
    DONT_WAIT = 2


# Module-level logger.
logger = logging.getLogger(__name__)

class TkWindow:
    """The Tk equivalent to a :class:`GWindow`."""
    def __init__(self, root, width, height, parent):
        self._parent = parent

        self._master = tk.Toplevel(root)
        self._closed = False
        self._master.protocol("WM_DELETE_WINDOW", self._close)
        self._master.resizable(width=False, height=False)  # Disable resizing by default.

        # Raise the master to be the top window.
        self._master.wm_attributes("-topmost", 1)  # TODO(sredmond): Is this really necessary?
        self._master.lift()
        self._master.focus_force()

        # TODO(sredmond): On macOS, multiple backends might race to set the process-level menu bar.
        setup_menubar(self._master)

        self._frame = tk.Frame(self._master)  #, bd=2, bg='red'

        self._canvas = tk.Canvas(self._frame, width=width, height=height, highlightthickness=0, bd=0)
        self._canvas.pack(fill=tk.BOTH, expand=True)

        self._frame.pack(fill=tk.BOTH, expand=True)
        self._frame.update()

        self._master.update()

        # Empty side regions for interaction. Their ordering and layout depends
        # on order of construction, so start them off empty.
        self._top = None
        self._bottom = None
        self._left = None
        self._right = None

    @property
    def canvas(self):
        return self._canvas

    @property
    def top(self):
        """Get the top bar for interactors, creating it if needed."""
        if not self._top:
            self._top = tk.Frame(self._master)
            self._top.pack(fill=tk.X, side=tk.TOP)
            self._frame.pack_forget()
            self._frame.pack(fill=tk.BOTH, expand=True)
        return self._top

    @property
    def bottom(self):
        """Get the bottom bar for interactors, creating it if needed."""
        if not self._bottom:
            self._bottom = tk.Frame(self._master)
            self._bottom.pack(fill=tk.X, side=tk.BOTTOM)
            self._frame.pack_forget()
            self._frame.pack(fill=tk.BOTH, expand=True)
        return self._bottom

    @property
    def left(self):
        """Get the left bar for interactors, creating it if needed."""
        if not self._left:
            self._left = tk.Frame(self._master)
            self._left.pack(fill=tk.Y, side=tk.LEFT)
            self._frame.pack_forget()
            self._frame.pack(fill=tk.BOTH, expand=True)
        return self._left

    @property
    def right(self):
        """Get the right bar for interactors, creating it if needed."""
        if not self._right:
            self._right = tk.Frame(self._master)
            self._right.pack(fill=tk.Y, side=tk.LEFT)
            self._frame.pack_forget()
            self._frame.pack(fill=tk.BOTH, expand=True)
        return self._right

    def clear(self):
        # Delete all canvas elements and all interactors, but leave the canvas and interactor regions in place."""
        self.clear_canvas()
        if self._top:
            for child in self._top.children:
                child.destroy()
        if self._bottom:
            for child in self._bottom.children:
                child.destroy()
        if self._left:
            for child in self._left.children:
                child.destroy()
        if self._right:
            for child in self._right.children:
                child.destroy()

    def clear_canvas(self):
        # Delete all canvas elements, but leave the canvas (and all interactor regions) in place.
        self.canvas.delete('all')

    def _close(self):
        if self._closed: return
        self._closed = True

        self._master.destroy()
        # TODO(sredmond): Consider autoflushing like Zelle.
        self._parent._remove_tkwin(self)  # Tell the parent that we have closed.


class TkBackend(GraphicsBackendBase):
    def __init__(self):
        self._root = tk.Tk()  # A wrapper around a new Tcl interpreter.
        self._root.withdraw()  # Removes the window from the screen (without destroying it).

        atexit.register(self._root.mainloop)  # TODO(sredmond): For debugging only.
        self._windows = []  # TODO(sredmond): Use winfo_children().

    def _update_active_window(self, window):
        # Optimization: Don't mess with the windows when there's only one.
        if len(self._windows) == 1: return

        window._master.lift()

        # Move the window to top of the stack.
        self._windows.remove(window)
        self._windows.append(window)

    def _remove_tkwin(self, window):
        if not window._closed:
            window._close()

        try:
            self._windows.remove(window)
        except ValueError:
            pass

        if not self._windows:
            self._shutdown()

    def _shutdown(self):
        self._root.destroy()

    ######################
    # GWindow lifecycle. #
    ######################
    def gwindow_constructor(self, gwindow, width, height, top_compound, visible=True):
        gwindow._tkwin = TkWindow(self._root, width, height, parent=self)
        gwindow._tkwin._gwindow = gwindow  # Circular reference so a TkWindow knows its originating GWindow.

        self._windows.append(gwindow._tkwin)

        # HACK: Get nice titles built in.
        self.gwindow_set_window_title(gwindow, gwindow.title)

    def gwindow_close(self, gwindow):
        self._remove_tkwin(gwindow._tkwin)

    def gwindow_delete(self, gwindow):
        self._remove_tkwin(gwindow._tkwin)

    def gwindow_exit_graphics(self):
        self._remove_tkwin(gwindow._tkwin)

    def gwindow_set_exit_on_close(self, gwindow, exit_on_close): pass

    ####################
    # GWindow drawing. #
    ####################
    def gwindow_clear(self, gwindow):
        self._update_active_window(gwindow._tkwin)
        gwindow._tkwin.clear()

    def gwindow_clear_canvas(self, gwindow):
        self._update_active_window(gwindow._tkwin)
        gwindow._tkwin.clear_canvas()

    def gwindow_repaint(self, gwindow):
        # Update any unresolved tasks.
        gwindow._tkwin._master.update_idletasks()

    def gwindow_draw(self, gwindow, gobject): pass

    ####################
    # GWindow drawing. #
    ####################
    def gwindow_request_focus(self, gwindow):
        self._update_active_window(gwindow._tkwin)
        gwindow._tkwin._master.focus_force()

    def gwindow_set_visible(self, gwindow, flag):
        self._update_active_window(gwindow._tkwin)
        if flag:  # Show the window.
            gwindow._tkwin._master.deiconify()
        else:  # Show the window.
            gwindow._tkwin._master.withdraw()

    def gwindow_set_window_title(self, gwindow, title):
        self._update_active_window(gwindow._tkwin)
        gwindow._tkwin._master.title(title)

    def gwindow_get_width(self):
        self._update_active_window(gwindow._tkwin)
        return gwindow._tkwin._master.geometry()[0]

    def gwindow_get_height(self):
        self._update_active_window(gwindow._tkwin)
        return gwindow._tkwin._master.geometry()[1]

    ######################
    # GWindow alignment. #
    ######################
    def gwindow_add_to_region(self, gwindow, gobject, region):
        from campy.graphics.gwindow import Region
        if region == Region.NORTH:
            self._ginteractor_add(gobject, gwindow._tkwin.top)


    def gwindow_remove_from_region(self, gwindow, gobject, region): pass
    def gwindow_set_region_alignment(self, gwindow, region, align): pass

    ##############################
    # Shared GObject operations. #
    ##############################
    def gobject_set_location(self, gobject, x, y):
        if not hasattr(gobject, '_tkid'): return

        tkid = gobject._tkid
        win = gobject._tkwin

        coords = win.canvas.coords(tkid)
        win.canvas.move(tkid, x - coords[0], y - coords[1])
        win._master.update_idletasks()

    def gobject_set_filled(self, gobject, flag):
        from campy.graphics.gobjects import GArc
        if not hasattr(gobject, '_tkid'): return

        tkid = gobject._tkid
        win = gobject._tkwin
        if flag:
            win.canvas.itemconfig(tkid, fill=gobject.fill_color.hex)
            if isinstance(object, GArc):
                win.canvas.itemconfig(tkid, style=tk.PIESLICE)
        else:
            win.canvas.itemconfig(tkid, fill='')
            if isinstance(object, GArc):
                self.itemconfig(tkid, style=tkinter.ARC)

        win._master.update_idletasks()

    def gobject_remove(self, gobject):
        if not hasattr(gobject, '_tkid'): return
        tkid = gobject._tkid
        win = gobject._tkwin

        win.canvas.delete(tkid)
        delattr(gobject, '_tkid')
        delattr(gobject, '_tkwin')

        win._master.update_idletasks()

    def gobject_set_color(self, gobject, color):
        if not hasattr(gobject, '_tkid'): return
        tkid = gobject._tkid
        win = gobject._tkwin

        # Awkward import.
        from campy.graphics.gobjects import GLabel, GLine
        if not isinstance(gobject, GLabel) and not isinstance(gobject, GLine):
            win.canvas.itemconfig(tkid, outline=color.hex)
        else:
            # GLabels and GLines are special because their "color" is actually a fill color.
            win.canvas.itemconfig(tkid, fill=color.hex)

        win._master.update_idletasks()

    def gobject_set_fill_color(self, gobject, color):
        if not hasattr(gobject, '_tkid'): return
        tkid = gobject._tkid
        win = gobject._tkwin

        win.canvas.itemconfig(tkid, fill=color.hex)

        win._master.update_idletasks()

    def gobject_send_forward(self, gobject): pass
    def gobject_send_to_front(self, gobject):
        if not hasattr(gobject, '_tkid'): return
        tkid = gobject._tkid
        win = gobject._tkwin

        win.canvas.tag_raise(tkid)

        win._master.update_idletasks()

    def gobject_send_backward(self, gobject): pass
    def gobject_send_to_back(self, gobject):
        if not hasattr(gobject, '_tkid'): return
        tkid = gobject._tkid
        win = gobject._tkwin

        win.canvas.tag_lower(tkid)

        win._master.update_idletasks()

    def gobject_set_size(self, gobject, width, height): pass
    def gobject_get_size(self, gobject):
        if not hasattr(gobject, '_tkid'): return
        tkid = gobject._tkid
        win = gobject._tkwin

        x0, y0, x1, y1 = win.canvas.bbox(tkid)
        return x1 - x0, y1 - y0

    def gobject_get_bounds(self, gobject): pass
    def gobject_set_line_width(self, gobject, line_width): pass
    def gobject_contains(self, gobject, x, y): pass
    def gobject_set_visible(self, gobject, flag):
        if not hasattr(gobject, '_tkid'): return
        tkid = gobject._tkid
        win = gobject._tkwin

        if not flag:
            win.canvas.itemconfig(tkid, state=tk.HIDDEN)
        else:
            win.canvas.itemconfig(tkid, state=tk.NORMAL)

    def gobject_scale(self, gobject, sx, sy): pass
    def gobject_rotate(self, gobject, theta): pass

    ########################
    # Rectangular regions. #
    ########################
    def grect_constructor(self, grect):
        if hasattr(grect, '_tkwin'):
            return

        win = self._windows[-1]
        grect._tkwin = win

        grect._tkid = win.canvas.create_rectangle(
            grect.x, grect.y, grect.x + grect.width, grect.y + grect.height,
            outline=grect.color.hex, fill=grect.fill_color.hex if grect.filled else '',
            state=tk.NORMAL if grect.visible else tk.HIDDEN)
        win._master.update_idletasks()

    def groundrect_constructor(self, gobject, width, height, corner): pass
    def g3drect_constructor(self, gobject, width, height, raised): pass
    def g3drect_set_raised(self, gobject, raised): pass

    #######################
    # Elliptical regions. #
    #######################
    def goval_constructor(self, goval):
        if hasattr(goval, '_tkwin'):
            return

        win = self._windows[-1]
        goval._tkwin = win

        goval._tkid = win.canvas.create_oval(
            goval.x, goval.y, goval.x + goval.width, goval.y + goval.height,
            outline=goval.color.hex, fill=goval.fill_color.hex if goval.filled else '',
            state=tk.NORMAL if goval.visible else tk.HIDDEN)

        win._master.update_idletasks()

    def garc_constructor(self, garc):
        if hasattr(garc, '_tkwin'):
            return

        win = self._windows[-1]
        garc._tkwin = win

        garc._tkid = win.canvas.create_arc(
            garc.x, garc.y, garc.x + garc.width, garc.y + garc.height,
            start=garc.start, extent=garc.sweep,
            outline=garc.color.hex, fill=garc.fill_color.hex if garc.filled else '',
            style=tk.PIESLICE if garc.filled else tk.ARC,
            state=tk.NORMAL if garc.visible else tk.HIDDEN)

        win._master.update_idletasks()

    def garc_set_start_angle(self, garc, angle):
        if not hasattr(garc, '_tkid'): return
        tkid = garc._tkid
        win = garc._tkwin

        win.canvas.itemconfig(tkid, start=angle)

        win._master.update_idletasks()

    def garc_set_sweep_angle(self, garc, angle):
        if not hasattr(garc, '_tkid'): return
        tkid = garc._tkid
        win = garc._tkwin

        win.canvas.itemconfig(tkid, extent=angle)

        win._master.update_idletasks()

    def garc_set_frame_rectangle(self, garc, x, y, width, height): pass


    ##########
    # GLines #
    ##########
    def gline_constructor(self, gline):
        if hasattr(gline, '_tkwin'):
            return

        win = self._windows[-1]
        gline._tkwin = win

        gline._tkid = win.canvas.create_line(
            gline.start.x, gline.start.y, gline.end.x, gline.end.y,
            fill=gline.color.hex,
            state=tk.NORMAL if gline.visible else tk.HIDDEN)

        win._master.update_idletasks()

    def gline_set_start_point(self, gline, x, y):
        if not hasattr(gline, '_tkid'): return
        tkid = gline._tkid
        win = gline._tkwin

        win.canvas.coords(tkid, x, y, gline.end.x, gline.end.y,)

        win._master.update_idletasks()

    def gline_set_end_point(self, gline, x, y):
        if not hasattr(gline, '_tkid'): return
        tkid = gline._tkid
        win = gline._tkwin

        win.canvas.coords(tkid, gline.start.x, gline.start.y, x, y)

        win._master.update_idletasks()

    #########
    # Fonts #
    #########
    # See: https://www.astro.princeton.edu/~rhl/Tcl-Tk_docs/tk/font.n.html

    def gfont_default_attributes(self):
        # Resolves to the platform-specific default.
        font = tkfont.nametofont('TkDefaultFont')
        return font.config()

    def gfont_attributes_from_system_name(self, font_name):
        # Attempt to load the font with the given name.
        font = tkfont.nametofont(font_name)
        return font.config()

    def gfont_get_font_metrics(self, gfont):
        font = tkfont.Font(family=gfont.family, size=gfont.size,
                           weight=tkfont.BOLD if gfont.weight else tkfont.NORMAL,
                           slant=tkfont.ITALIC if gfont.slant else tkfont.ROMAN)
        if not hasattr(gfont, '_tkfont'):
            gfont._tkfont = font
        return font.metrics()

    def gfont_measure_text_width(self, gfont, text):
        if not hasattr(gfont, '_tkfont'):
            gfont._tkfont = tkfont.Font(family=gfont.family, size=gfont.size,
                            weight=tkfont.BOLD if gfont.weight else tkfont.NORMAL,
                            slant=tkfont.ITALIC if gfont.slant else tkfont.ROMAN)
        font = gfont._tkfont
        return font.measure(text)

    ##########
    # Labels #
    ##########
    def glabel_constructor(self, glabel):
        if hasattr(glabel, '_tkwin'):
            return

        win = self._windows[-1]
        glabel._tkwin = win

        # TODO(sredmond): Document that we're putting the anchor at the NW corner.
        # TODO(sredmond): Respect the font that's been set.
        glabel._tkid = win.canvas.create_text(
            glabel.x, glabel.y,
            text=glabel.text,
            fill=glabel.color.hex, anchor=tk.SW,
            state=tk.NORMAL if glabel.visible else tk.HIDDEN)

        win._master.update_idletasks()

    def glabel_set_font(self, glabel, gfont):
        if not hasattr(glabel, '_tkid'): return

        if not hasattr(gfont, '_tkfont'):
            gfont._tkfont = tkfont.Font(family=gfont.family, size=gfont.size,
                            weight=tkfont.BOLD if gfont.weight else tkfont.NORMAL,
                            slant=tkfont.ITALIC if gfont.slant else tkfont.ROMAN)
        font = gfont._tkfont
        tkid = glabel._tkid
        win = glabel._tkwin
        win.canvas.itemconfig(tkid, font=font)

    def glabel_set_label(self, glabel, text):
        if not hasattr(glabel, '_tkid'): return
        tkid = glabel._tkid
        win = glabel._tkwin

        win.canvas.itemconfig(tkid, text=text)

    def glabel_get_font_ascent(self, glabel): pass
    def glabel_get_font_descent(self, glabel): pass
    def glabel_get_size(self, glabel):
        # TODO(sredmond): This is currently broken.
        if not hasattr(glabel, '_tkid'): return 0, 0
        tkid = glabel._tkid
        win = glabel._tkwin

        x0, y0, x1, y1 = win.canvas.bbox(tkid)
        return x1 - x0, y1 - y0

    # Polygons
    def gpolygon_constructor(self, gpolygon):
        if hasattr(gpolygon, '_tkwin'):
            return

        win = self._windows[-1]
        gpolygon._tkwin = win

        coords = sum(((v.x + gpolygon.x, v.y + gpolygon.y) for v in gpolygon.vertices), ())
        gpolygon._tkid = win.canvas.create_polygon(coords,  # Not the fastest, but it'll do
            outline=gpolygon.color.hex, fill=gpolygon.fill_color.hex if gpolygon.filled else '',
            state=tk.NORMAL if gpolygon.visible else tk.HIDDEN)

        win._master.update_idletasks()

    def gpolygon_add_vertex(self, gpolygon, x, y):
        if not hasattr(gpolygon, '_tkid'): return
        tkid = gpolygon._tkid
        win = gpolygon._tkwin

        win.canvas.coords(tkid, x, y, gpolygon.end.x, gpolygon.end.y,)

        win._master.update_idletasks()

    ##########
    # Events #
    ##########
    def set_action_command(self, gobject, cmd): pass
    def get_next_event(self, mask): pass
    def wait_for_event(self, mask): pass

    def event_add_keypress_handler(self, event, handler): pass
    def event_generate_keypress(self, event): pass

    def _wrap_mouse_event(self, event, window, event_type):
        from campy.gui.events.mouse import GMouseEvent
        # TODO(sredmond): As written, this joins the TkWindow, not the GWindow, to this event.
        return GMouseEvent(event_type=event_type, gwindow=window._gwindow, x=event.x, y=event.y)

    def event_add_mouse_handler(self, event, handler):
        from campy.gui.events.mouse import MouseEventType
        if not self._windows:
            logger.warning('Refusing to add a mouse listener before any windows are created.')
            return

        win = self._windows[-1]

        if event == MouseEventType.MOUSE_CLICKED:
            win._master.bind('<Button-1>', lambda cb_event: handler(self._wrap_mouse_event(cb_event, win, event)))
        elif event == MouseEventType.MOUSE_RELEASED:
            win._master.bind('<ButtonRelease-1>', lambda cb_event: handler(self._wrap_mouse_event(cb_event, win, event)))
        elif event == MouseEventType.MOUSE_MOVED:
            win._master.bind('<Motion>', lambda cb_event: handler(self._wrap_mouse_event(cb_event, win, event)))
        elif event == MouseEventType.MOUSE_DRAGGED:
            win._master.bind('<B1-Motion>', lambda cb_event: handler(self._wrap_mouse_event(cb_event, win, event)))
        else:
            logger.warning('Unrecognized event type: {}. Quietly passing.'.format(event))

    def event_generate_mouse(self, event): pass

    def event_pump_one(self):
        # Forcibly process queued tasks, but don't process newly queued ones.
        self._root.update_idletasks()
        self._root.dooneevent(DONT_WAIT)

    # TODO(sredmond): Rename these backend events for consistency.
    def timer_pause(self, event): pass
    def timer_schedule(self, function, delay_ms):
        self._root.after(delay_ms, function)


    ###############
    # Interactors #
    ###############
    def _ginteractor_add(self, gint, frame):
        from campy.gui.ginteractors import GButton
        if isinstance(gint, GButton):
            # TODO(sredmond): Wrap up a GActionEvent on the Tk side to supply.
            gint._tkobj = tk.Button(frame, text=gint.label, command=gint.click,
            state=tk.NORMAL if not gint.disabled else tk.DISABLED)
        gint._tkobj.pack()

        frame.update_idletasks()

    def gbutton_constructor(self, gbutton):
        if hasattr(gbutton, '_tkwin'):
            return

        win = self._windows[-1]
        gbutton._tkwin = win

    def gbutton_set_label(self, gbutton):
        if not hasattr(gbutton, '_tkobj'): return
        gbutton._tkobj.config(text=gbutton.label)

    def gbutton_set_disabled(self, gbutton):
        if not hasattr(gbutton, '_tkobj'): return
        gbutton._tkobj.config(state=tk.NORMAL if not gbutton.disabled else tk.DISABLED)

    def gcheckbox_constructor(self, gcheckbox):
        if hasattr(gcheckbox, '_tkwin'):
            return

        win = self._windows[-1]
        gcheckbox._tkwin = win

        # TODO(sredmond): Wrap up a GActionEvent on the Tk side to supply.
        var = tkinter.IntVar()
        gcheckbox._tkobj = tk.Checkbutton(win._master, text=gcheckbox.label, command=gcheckbutton.select,
            variable=var,
            state=tk.NORMAL if not gcheckbox.disabled else tk.DISABLED)
        gcheckbox._tkobj.var = var
        gcheckbox._tkobj.pack()

        win._master.update_idletasks()


    def gcheckbox_is_selected(self, gcheckbox):
        return bool(gcheckbox._tkobj.var.get())

    def gcheckbox_set_selected(self, gcheckbox, state):
        return gcheckbox._tkobj.var.set(1 if state else 0)

    def gslider_constructor(self, gslider, min, max, value): pass
    def gslider_get_value(self, gslider): pass
    def gslider_set_value(self, gslider, value): pass
    def gtextfield_constructor(self, gtextfield, num_chars): pass
    def gtextfield_get_text(self, gtextfield): pass
    def gtextfield_set_text(self, gtextfield, str): pass
    def gchooser_constructor(self, gchooser): pass
    def gchooser_add_item(self, gchooser, item): pass
    def gchooser_get_selected_item(self, gchooser): pass
    def gchooser_set_selected_item(self, gchooser, item): pass

    ###########
    # Dialogs #
    ###########
    # TODO(sredmond): Make these dialogs steal focus.
    def gfilechooser_show_open_dialog(self, current_dir, file_filter):
        logger.debug('Ignoring file_filter argument to gfilechooser_show_open_dialog.')
        parent = None
        if self._windows:
            parent=self._windows[-1]
        return tkfiledialog.askopenfilename(initialdir=current_dir, title='Select File to Open', parent=parent) or None

    def gfilechooser_show_save_dialog(self, current_dir, file_filter):
        logger.debug('Ignoring file_filter argument to gfilechooser_show_save_dialog.')
        parent = None
        if self._windows:
            parent=self._windows[-1]
        return tkfiledialog.asksaveasfilename(initialdir=current_dir, title='Select File to Save', parent=parent) or None

    def goptionpane_show_confirm_dialog(self, message, title, confirm_type):
        from campy.graphics.goptionpane import ConfirmType

        if confirm_type == ConfirmType.YES_NO:
            return tkmessagebox.askyesno(title, message)
        elif confirm_type == ConfirmType.YES_NO_CANCEL:
            return tkmessagebox.askyesnocancel(title, message)
        elif confirm_type == ConfirmType.OK_CANCEL:
            return tkmessagebox.askokcancel(title, message)
        else:
            logger.debug('Unrecognized confirm_type {!r}'.format(confirm_type))

    def goptionpane_show_input_dialog(self, message, title):
        return tksimpledialog.askstring(title, message, parent=self._root)

    def goptionpane_show_message_dialog(self, message, title, message_type):
        from campy.graphics.goptionpane import MessageType

        # TODO(sredmond): The icons aren't appearing correctly.
        if message_type == MessageType.ERROR:
            tkmessagebox.showerror(title, message, icon=tkmessagebox.ERROR)
        elif message_type == MessageType.INFORMATION:
            tkmessagebox.showinfo(title, message, icon=tkmessagebox.INFO)
        elif message_type == MessageType.WARNING:
            tkmessagebox.showwarning(title, message, icon=tkmessagebox.WARNING)
        elif message_type == MessageType.QUESTION:
            tkmessagebox.showinfo(title, message, icon=tkmessagebox.QUESTION)
        elif message_type == MessageType.PLAIN:
            tkmessagebox.showinfo(title, message)
        else:
            logger.debug('Unrecognized message_type {!r}'.format(message_type))

    def goptionpane_show_option_dialog(self, message, title, options, initially_selected): pass
    def goptionpane_show_text_file_dialog(self, message, title, rows, cols): pass

if __name__ == '__main__':
    # Quick tests.
    from campy.graphics.gwindow import GWindow
    from campy.graphics.gobjects import GRect, GPolygon
    from campy.graphics.gfilechooser import show_open_dialog, show_save_dialog
    from campy.graphics.goptionpane import *
    from campy.graphics.gtypes import *
    from campy.gui.interactors import *

    import math

    print('{!r}'.format(show_open_dialog()))
    print('{!r}'.format(show_save_dialog()))


    window = GWindow()
    rect = GRect(100, 200, x=50, y=60)
    window.add(rect)
    rect.location = 300, 300
    button = GButton('Button')
    window.add(button)

    # Add a polygon.
    edge_length = 75
    stop_sign = GPolygon()
    start = GPoint(-edge_length / 2, edge_length / 2 + edge_length / math.sqrt(2.0))
    stop_sign.add_vertex(start)
    for edge in range(8):
        stop_sign.add_polar_edge(edge_length, 45*edge)
    stop_sign.filled = True
    stop_sign.color = "BLACK"
    stop_sign.fill_color = "RED"
    window.add(stop_sign, window.width / 2, window.height / 2)




import inspect as _inspect
import functools as _functools

## Decorators
def _bind_args(function, *args, **kwargs):
    """Returns an map from the names of function's arguments to values given by *args and **kwargs

    This is more or less an implementation of Python argument bind semantics, but it's not super accurate
    ¯\_(ツ)_/¯
        For example, it doesn't resolve any closure elements or anything, because ahh that's awful

    First, positional arguments are bound

    If you're a student reading this, you can ignore this implementation.

    Pre: *args and **kwargs represent valid parameters
    """
    argspec = _inspect.getfullargspec(function)
    sig = _inspect.Signature.from_function(function)
    ba = sig.bind(*args, **kwargs)
    bindings = ba.arguments.copy()
    # default values for keyword arguments
    if argspec.defaults:
        for var_name, default_value in zip(reversed(argspec.args), reversed(argspec.defaults)):
            if var_name not in bindings:
                bindings[var_name] = default_value
    # default values for keyword-only argument
    if argspec.kwonlydefaults:
        for var_name, default_value in argspec.kwonlydefaults.items():
            if var_name not in bindings:
                bindings[var_name] = default_value
    if argspec.varargs and argspec.varargs not in bindings:
        bindings[argspec.varargs] = tuple()
    if argspec.varkw and argspec.varkw not in bindings:
        bindings[argspec.varkw] = dict()
    return bindings

def print_args(function):
    """Decorate the given function to print out it's arguments and return val if not None

    """
    @_functools.wraps(function)
    def wrapper(*args, **kwargs):
        bound_arguments = _bind_args(function, *args, **kwargs)
        print("{name}({call})".format(
            name=function.__name__,
            call=', '.join("{}={}".format(arg, val) for arg, val in bound_arguments.items())
        ))
        retval = function(*args, **kwargs)
        if retval is not None:
            print("(return) {!r}".format(retval))
        return retval
    return wrapper

# def test_print_args():
#     @print_args
#     def all_together(x, y, z=1, *nums, indent=True, spaces=4, **options):
#         """The all_together function from Lab 3."""
#         pass

#     # all_together(2)
#     all_together(2, 5, 7, 8, indent=False)
#     all_together(2, 5, 7, 6, indent=None)
#     all_together()
#     # all_together(indent=True, 3, 4, 5)
#     # all_together(**{'indent': False}, scope='maximum')
#     all_together(dict(x=0, y=1), *range(10))
#     # all_together(**dict(x=0, y=1), *range(10))
#     # all_together(*range(10), **dict(x=0, y=1))
#     all_together([1, 2], {3:4})
#     all_together(8, 9, 10, *[2, 4, 6], x=7, spaces=0, **{'a':5, 'b':'x'})
#     all_together(8, 9, 10, *[2, 4, 6], spaces=0, **{'a':[4,5], 'b':'x'})
#     # all_together(8, 9, *[2, 4, 6], *dict(z=1), spaces=0, **{'a':[4,5], 'b':'x'})


def cache(function):
    function._cache = {}
    @_functools.wraps(function)
    def wrapper(*args, **kwargs):
        key = (args, tuple(kwargs.items()))
        if key in function._cache:
            return function._cache[key]
        retval = function(*args, **kwargs)
        function._cache[key] = retval
        return retval
    return wrapper

# @cache
# def fib(n):
#     return fib(n-1) + fib(n-2) if n > 2 else 1

def cache_challenge(max_size=None, eviction_policy='LRU'):
    assert eviction_policy in ['LRU', 'MRU', 'random']
    def decorator(function):
        function._cache = collections.OrderedDict()
        @_functools.wraps(function)
        def wrapper(*args, **kwargs):
            key = (args, tuple(kwargs.items()))
            if key in function._cache:
                # Before accessing this element, move it to the MRU side
                # of the list
                function._cache.move_to_end(key)
                return function._cache[key]
            retval = function(*args, **kwargs)

            # Check for eviction
            if max_size and len(function._cache) == max_size:
                if eviction_policy == 'LRU':
                    function._cache.popitem(last=False)
                elif eviction_policy == 'MRU':
                    function._cache.popitem(last=True)
                else:
                    randkey = random.choice(list(function._cache.keys()))
                    function._cache.pop(randkey)
            # Now that we know there's space, insert the element
            function._cache[key] = retval
            return retval
        return wrapper
    return decorator

# @cache_challenge(max_size=16, eviction_policy='LRU')
# def fib(n):
#     return fib(n-1) + fib(n-2) if n > 2 else 1


def enforce_types(function):
    expected = function.__annotations__
    if not expected:
        return function
    assert(all(map(lambda exp: type(exp) == type, expected.values())))
    @_functools.wraps(function)
    def wrapper(*args, **kwargs):
        bound_arguments = _bind_args(function, *args, **kwargs)
        for arg, val in bound_arguments.items():
            if arg in expected and not isinstance(val, expected[arg]):
                print("(Bad Argument Type!) argument '{arg}={val}': expected {exp}, received {r}".format(
                    arg=arg,
                    val=val,
                    exp=expected[arg],
                    r=type(val)
                ))

        retval = function(*args, **kwargs)

        # Check the return value
        if 'return' in expected and not isinstance(retval, expected['return']):
            print("(Bad Return Value!) return '{ret}': expected {exp}, received {r}".format(
                ret=retval,
                exp=expected['return'],
                r=type(retval)
            ))
        return retval
    return wrapper

# def test_enforce_types():
#     @enforce_types
#     def foo(a: int, b: str) -> bool:
#         if a == -1:
#             return 'Gotcha!'
#         return b[a] == 'X'

#     try:
#         foo(3, 'XYZXYZ')  # => True
#         foo(2, 'python')  # => False
#         foo(1, 4)  # prints "(Bad Argument Type!) argument b=4: expected <class 'str'>, received <class 'int'>" and then crashes
#         foo(-1, '')  # prints "(Bad Return Value!) return Gotcha!: expected <class 'bool'>, received <class 'str'>" and returns "Gotcha!"
#     except TypeError:
#         pass



def enforce_types_challenge(severity=1):
    assert severity in [0, 1, 2]
    if severity == 0:
        # Return a no-op decorator
        return lambda function: function

    def message(msg):
        if severity == 1:
            print(msg)
        else:
            raise TypeError(msg)

    def decorator(function):
        expected = function.__annotations__
        if not expected:
            return function
        assert(all(map(lambda exp: type(exp) == type, expected.values())))

        @_functools.wraps(function)
        def wrapper(*args, **kwargs):
            bound_arguments = _bind_args(function, *args, **kwargs)
            for arg, val in bound_arguments.items():
                if arg in expected and not isinstance(val, expected[arg]):
                    msg("(Bad Argument Type!) argument '{arg}={val}': expected {exp}, received {r}".format(
                        arg=arg,
                        val=val,
                        exp=expected[arg],
                        r=type(val)
                    ))

            retval = function(*args, **kwargs)

            # Check the return value
            if 'return' in expected and not isinstance(retval, expected['return']):
                msg("(Bad Return Value!) return '{ret}': expected {exp}, received {r}".format(
                    ret=retval,
                    exp=expected['return'],
                    r=type(retval)
                ))
            return retval
        return wrapper
    return decorator

def stylist(old, new, details):
    def stylist_decorator(function):
        @_functools.wraps(function)
        def stylist_wrapper(*args, **kwargs):
            print('Access to {0} has been stylistically replaced.'.format(function.__name__))
            print(old)
            print(new)
            return function(*args, **kwargs)
        return stylist_wrapper
    return stylist_decorator


# Example code.
# @stylist("GOval.getWidth()", "GOval.width", "In Python, you can just access this attribute directly.")
# def getWidth():
# print('I am a getWidth function.')



"""Friendly fonts.

A :class:`GFont` provides a convenient way to define fonts.

To get the current platform's default font::

    font = GFont.default()

To get a font from a text description in the form FAMILY-STYLE or
FAMILY-STYLE-MODIFIERS::

    font1 = GFont.parse('Helvetica-12')
    font2 = GFont.parse('Times New Roman-24-bold')
    font3 = GFont.parse('Verdana-11-bold-italic')

For finer grained control, use the constructor directly.

    font1 = GFont.parse('Helvetica', 12)
    font2 = GFont.parse('Times New Roman', 24, weight=True)
    font3 = GFont.parse('Verdana', 11, weight=True, slant=True)

To access a system font by name, such as the macOS alert header::

    font = GFont.system('systemAlertHeaderFont')

Note that a given system font may not be accessible on all platforms.

Once you have a :class:`GFont`, you can access its family, size, weight, and
slant. You can also access font metrics directly, such as the font's ascent,
the font's descent, the font's line spacing, and whether the font is
fixed-width. These are properties that apply to the font as a whole, and not
to any particular text rendered in this font.

To get the width of a particular string, use :func:`GFont.measure`::

    font = GFont.default()
    width = font.measure('Hello World!')
    print(width)  # => 77
"""
import campy.private.platform as _platform

import logging

# Module-level logger.
logger = logging.getLogger(__name__)


class GFont:
    # TODO(sredmond): Support underline and overstrike.

    # TODO(sredmond): Think harder about which of these should be required and
    # which should be optional.
    def __init__(self, family, size, weight=False, slant=False):
        self._family = family
        self._size = size
        self._weight = weight
        self._slant = slant

        metrics = _platform.Platform().gfont_get_font_metrics(self)
        self._ascent = metrics['ascent']
        self._descent = metrics['ascent']
        self._linespace = metrics['ascent']
        # TODO(sredmond): Push the actual value of 1 back down into the platform abstraction.
        self._fixed = metrics['fixed'] == '1'

    @classmethod
    def parse(cls, description):
        """Parse a description into a font.

        A description can be either FAMILY-SIZE or FAMILY-SIZE-MODIFIERS, where
        FAMILY is the name of a font family available on the current system,
        SIZE is a positive integer representing the font size (in points, not
        in pixels), and MODIFIERS is either the string "bold", "bold-italic",
        or "italic". Examples of valid font descriptions include::


            Helvetica-12
            Dialog-13
            Courier-18
            Times New Roman-24-bold
            Times-16-italic
            Verdana-11-bold-italic

        This method returns None if no font could be parsed.
        """
        # TODO(sredmond): Redesign the allowed description scheme to handle a
        # broader set of descriptions. Notably, fonts with dashes in their
        # names are currently impossible to specify. I don't know of any fonts
        # with that property, but it's still an unnecessary restriction.
        pieces = description.split('-')
        if len(pieces) < 2 or len(pieces) > 4:
            logger.warning('Unable to parse font description {!r} into an appropriate number of pieces.'.format(description))
            return
        family = pieces[0]
        size = int(pieces[1])
        weight = 'bold' in pieces[2:]
        slant = 'italic' in pieces[2:]
        return cls(family, size, weight, slant)

    @classmethod
    def system(cls, font_name):
        logger.info('Loading a font from a system name may not be portable.')
        attributes = _platform.Platform().gfont_attributes_from_system_name(font_name)
        return cls(family=attributes['family'], size=attributes['size'],
                   weight=bool(attributes['weight']), slant=bool(attributes['slant']))

    # TODO(sredmond): Also support X Font Descriptors, which is something like:
    # -*-family-weight-slant-*--*-size-*-*-*-*-charset

    @classmethod
    def default(cls):
        attributes = _platform.Platform().gfont_default_attributes()
        return cls(family=attributes['family'], size=attributes['size'],
                   weight=bool(attributes['weight']), slant=bool(attributes['slant']))
    @property
    def family(self):
        return self._family

    @property
    def size(self):
        return self._size

    @property
    def weight(self):
        return self._weight

    @property
    def slant(self):
        return self._slant

    @property
    def ascent(self):
        return self._ascent

    @property
    def descent(self):
        return self._descent

    @property
    def linespace(self):
        return self._linespace

    @property
    def fixed(self):
        return self._fixed

    def measure(self, text):
        return _platform.Platform().gfont_measure_text_width(self, text)

    def __str__(self):
        """Implement `str(self)`."""
        out = '{family}-{size}'.format(family=self.family, size=self.size)
        if self.weight:
            out += '-bold'
        if self.slant:
            out += '-italic'
        return out

    def __repr__(self):
        """Implement `repr(self)`."""
        return 'GFont(family={family}, size={size}, weight={weight}, slant={slant})'.format(
            family=self.family,
            size=self.size,
            weight='bold' if self.weight else 'normal',
            slant='italic' if self.slant else 'roman'
        )



"""Set up and manage a menu bar."""
import tkinter as tk

def setup_menubar(master):
    """Create a menu bar and attach it to a tk.Tk() or tk.Toplevel().

    This menu bar looks like::

        MENU
        |-- File/
        |   |-- Save/
        |   |-- Exit/
        .-- Help/
            .-- About
    """
    menubar = tk.Menu(master)

    filemenu = tk.Menu(menubar, tearoff=False)
    filemenu.add_command(label="Save", command=lambda: print('Save Menu clicked.'))
    filemenu.add_separator()
    filemenu.add_command(label="Exit", command=master.quit)  # TODO(sredmond): Why not master.destroy?
    menubar.add_cascade(label="File", menu=filemenu)

    helpmenu = tk.Menu(menubar, tearoff=False)
    helpmenu.add_command(label="About...", command=lambda: print('The campy libraries were created by Sam Redmond.'))
    menubar.add_cascade(label="Help", menu=helpmenu)

    master.config(menu=menubar)



#!/usr/bin/env python3 -tt
"""
File: bitstream.cpp
-------------------
This file defines the ibitstream and obitstream classes.
These classes are patterned after (and, in fact, inherit from) the standard
io.BufferedReader and io.BufferedWriter classes.

The ibitstream and obitstream classes are basically the
same as the ordinary io.BufferedReader and io.BufferedWriter classes, but add the
functionality to read and write one bit at a time.

The idea is that you can substitute an ibitstream in place of an
io.BufferedReader and use the same operations (read, close, tell, etc.)
along with added member functions of readBit, rewind, and size.

Similarly, the obitstream can be used in place of io.BufferedWriter, and has
same operations (write, close, seek, etc.) along with additional
member functions writeBit and size.

# You can use ibitstream in any There are two subclasses of ibitstream: ifbitstream and istringbitstream,
# which are similar to the io. and istringstream classes.  The
# obitstream class similarly has ofbitstream and ostringbitstream as
# subclasses.

Note: in keeping with the naming conventions of the Python standard library,
readBit and writeBit have been renamed as readbit and writebit

Additionally, str() in ibitstream has been removed (doesn't make much sense anyway)
and for consistency with the standard library str in obitstream has been renamed getvalue()

Usage:

To use an ifbitstream::

    with ifbitstream(filename) as stream:
        bit = stream.readbit()
        stream.rewind()
        bit = stream.readbit()

To use an ofbitstream::

    with ofbitstream(filename) as stream:
        stream.writebit(0)
        ...

To use an ostringbitstream::

    with ostringbitstream() as stream:
        stream.writebit(0)
        ...
        stream.getvalue()  # => b'hello world'

To use an istringbitstream::

    with istringstream(b'hello world') as stream:
        bit = stream.readbit()
"""

import io as _io

"""
Constant: PSEUDO_EOF
A constant representing the PSEUDO_EOF marker that you will
write at the end of your Huffman-encoded file.
"""
PSEUDO_EOF = 256;

"""
Constant: NOT_A_CHAR
A constant representing an extended character that does not
actually hold a value.    When you are constructing your Huffman
encoding tree, you should set the characters in each internal
node (non-leaf) to this value to explicitly mark that they are not
being used.
"""
NOT_A_CHAR = 257;

NUM_BITS_IN_BYTE = 8

def get_nth_bit(pos, byte):
    return (byte >> (NUM_BITS_IN_BYTE - 1 - pos)) & 1

def set_nth_bit(pos, byte):
    return byte | 1 << (NUM_BITS_IN_BYTE - 1 - pos)

class ibitstream(_io.BufferedReader):
    def __init__(self, raw, buffer_size=_io.DEFAULT_BUFFER_SIZE):
        super().__init__(raw, buffer_size)
        self._fake = False
        self.pos = NUM_BITS_IN_BYTE
        self.current_byte = 0
        self.last_tell = 0

    def readbit(self):
        if self.closed:
            raise ValueError("ibitstream.readbit: Cannot read a bit from a stream that is not open.")
        if self._fake:  # Fake mode is used for autograding, and reads bytes as if they were bits
            bit = read(1)
            if bit == 0 or bit == ord('0'):
                return 0
            else:
                return 1
        else:
            # We consumed a whole byte, or the stream changed under us
            if self.pos == NUM_BITS_IN_BYTE or self.last_tell != self.tell():
                self.current_byte = self.read(1)
                if not self.current_byte:  # EOS
                    return PSEUDO_EOF
                self.current_byte = ord(self.current_byte)
                self.pos = 0
                self.last_tell = self.tell()
            result = get_nth_bit(self.pos, self.current_byte)
            self.pos += 1
            return result

    def rewind(self):
        if not self.seekable():
            raise _io.UnsupportedOperation()
        return self.seek(0) == 0

    def size(self):
        cur = self.tell()
        self.seek(0, _io.SEEK_END)
        end = self.tell()
        self.seek(cur, _io.SEEK_SET)
        return end * NUM_BITS_IN_BYTE


class obitstream(_io.BufferedWriter):
    def __init__(self, raw, buffer_size=_io.DEFAULT_BUFFER_SIZE, always_flush=True):
        super().__init__(raw, buffer_size)
        self._fake = False
        self.pos = NUM_BITS_IN_BYTE
        self.current_byte = 0
        self.last_tell = 0
        self.always_flush = always_flush

    def writebit(self, bit):
        if bit not in (0, 1):
            raise ValueError("obitstream.writebit: must pass an integer argument of 0 or 1. You passed the integer {}".format(bit))
        if self.closed:
            raise ValueError("obitstream.writebit: Cannot write a bit to a stream that is not open.")
        if self._fake:
            self.write(b'0' if bit == 0 else b'1')
            if self.always_flush:
                self.flush()
        else:
            # We wrote a whole byte, or the stream changed under us
            if self.pos == NUM_BITS_IN_BYTE or self.last_tell != self.tell():
                self.current_byte = 0
                self.pos = 0
            if bit:
                self.current_byte = set_nth_bit(self.pos, self.current_byte)
            if self.pos == 0 or bit:  # Write the first bit, or a change from 0 to 1
                if self.pos:
                    self.seek(-1, _io.SEEK_CUR)
                self.write(bytes([self.current_byte]))
                if self.always_flush:
                    self.flush()

            self.pos += 1
            self.last_tell = self.tell()

    def size(self):
        cur = self.tell()
        self.seek(0, _io.SEEK_END)
        end = self.tell()
        self.seek(cur, _io.SEEK_SET)
        return end * NUM_BITS_IN_BYTE

class ifbitstream(ibitstream):
    def __init__(self, filename):
        self.stream = _io.open(filename, 'rb')
        super().__init__(self.stream)


class ofbitstream(obitstream):
    def __init__(self, filename):
        self.stream = _io.open(filename, 'wb')
        super().__init__(self.stream)


class istringbitstream(ibitstream):
    def __init__(self, string):
        self.stream = _io.BytesIO(string)
        super().__init__(self.stream)

    def setvalue(self, string):
        view = self.stream.getbuffer()
        view[:] = string


class ostringbitstream(obitstream):
    def __init__(self):
        self.stream = _io.BytesIO()
        super().__init__(self.stream)

    def getvalue(self):
        return self.stream.getvalue()

__all__ = ['ibitstream', 'obitstream', 'ifbitstream', 'ofbitstream', 'istringbitstream', 'ostringbitstream']



#!/usr/bin/env python3 -tt
"""
File: observable.h
------------------
This file defines an abstract superclass named <code>Observable</code> that
allows objects to store lists of observers, which are other objects that are
notified when some part of the state of the observable object changes.
This is an example of the classic Observer/Observable design pattern.
"""
from abc import abstractmethod as _abstractmethod

class Observable():
    def __init__(self):
        self.observers = []

    def add_observer(self, observer):
        if observer not in self.observers:
            self.observers.append(observer)

    def notify_observers(self, *args, **kwargs):
        for observer in self.observers:
            observer.update(self, *args, **kwargs)

    def remove_observer(self, observer):
        if observer in self.observers:
            self.observers.remove(observer)

class Observer():
    @_abstractmethod
    def update(self, observed, *args, **kwargs):
        pass

__all__ = ['Observable', 'Observer']



"""
File: consolestreambuf.py
-------------------------
This file defines the <code>ConsoleStreamBuffer</code> class, which
represents a stream buffer that reads/writes to the Stanford graphical console
using a process pipe to a Java back-end process.
"""
import io as _io
import campy.private.platform as _platform

class ConsoleStreamBuffer:
    def input(self, prompt=''):
        _platform.Platform().put_console(prompt)
        result = _platform.Platform().get_line_console()
        if result.endswith('\n'):
            result = result[:-1]
        return result

    def print(self, *objects, sep='', end='\n', flush=False, stderr=False):
        buffer = _io.StringIO()
        __builtins__.print(*objects, sep=sep, end=end, file=buffer, flush=True)
        data = buffer.getvalue()
        _platform.Platform().put_console(data, stderr=stderr)


def test_consolestreambuffer():
    console = ConsoleStreamBuffer()
    console.print('Hello there! I echo text.\n')
    name = console.input('What is your name? ')
    console.print('Hello, {}'.format(name))
    while True:
        to_say = console.input('What should I say (hit ENTER to exit)? ')
        if not to_say: break
        console.print(to_say)

    console.print(5, 6, 7, sep='/', end=' :)', stderr=True)

if __name__ == '__main__':
    test_consolestreambuffer()



"""Graphical interactors.

This module provides five types of interactors:

- :class:`GButton`: A clickable button.
- :class:`GCheckBox`: A checkable box.
- :class:`GSlider`: A slider between numeric values.
- :class:`GChooser`: A list of selectable items.
- :class:`GTextField`: A single-line string entry.
"""

import campy.graphics.gobjects as gobjects
import campy.private.platform as _platform

from campy.graphics.gevents import *


class GInteractor(gobjects.GObject):
    """Superclass of all graphical interactors.

    In most applications, interactors will be added to one region in a
    :class:`GWindow`, but interactors can also be placed in specific positions
    in a :class:`GCompound` (or the :class:`GWindow`'s default top
    :class:`GCompound`) just like any other :class:`GObject`.
    """
    def __init__(self, command=''):
        """Initialize a :class:`GInteractor` with a command.

        The command is accessible as part of any :class:`GActionEvents` that
        this interactor generates.

        :param command: The action command for this interactor.
        """
        # TODO(sredmond): Consider stripping the concept of "action commands"
        # entirely, since it violates "There should be one way to do things."
        super().__init__()
        self._command = command

    @property
    def command(self):
        """Get or set the action command for this interactor."""
        return self._command

    @command.setter
    def command(self, command):
        self._command = command
        _platform.Platform().ginteractor_set_action_command(self, action_command)

    # TODO(sredmond): Provide a general mechanism for setting an interactor's size.
    # TODO(sredmond): Provide a general mechanism to get the location and bounds
    # of an interactor.


class GButton(GInteractor):
    """An onscreen, clickable button.

    The following program displays a button that, when pressed, generates the
    message: "Please do not press this button again!" (credit Douglas Adams)::

        window = GWindow()
        button = GButton('RED BUTTON')
        window.add_to_region(button, Region.NORTH)

        @button.onclick
        def chastise(event):
            event.button.label = 'Please do not press this button again!'
            event.button.disable()
    """
    def __init__(self, label):
        """Construct a GButton with the specified label.

        :param str label: A label for this GButton.
        """
        super().__init__()
        self._label = label
        self._disabled = False

        # Collection of
        self._listeners = []

        # Tell the underlying platform to construct this object.
        _platform.Platform().gbutton_constructor(self)

    ##############
    # PROPERTIES #
    ##############
    @property
    def label(self):
        """Get or set this :class:`GButton`:'s label."""
        return self._label

    @label.setter
    def label(self, label):
        self._label = label
        _platform.Platform().gbutton_set_label(self)

    @property
    def disabled(self):
        """Get or set whether this :class:`GButton`: is disabled."""
        return self._disabled

    @disabled.setter
    def disabled(self, disabled):
        self._disabled = disabled
        _platform.Platform().gbutton_set_disabled(self)

    # TODO(sredmond): Consider removing these convenience functions.
    def enable(self):
        """Enable this :class:`GButton`, meaning it can receive clicks."""
        self.disabled = False

    def disable(self):
        """Disable this :class:`GButton`, meaning it cannot receive clicks."""
        self.disabled = True

    #################
    # BUTTON EVENTS #
    #################
    def onclick(self, function):
        self._listeners.append(function)
        return function

    def add_click_handler(self, function):
        self._listeners.append(function)

    def remove_click_handler(self, function):
        try:
            self._listeners.remove(function)
            return True
        except ValueError:
            return False

    def _click(self, event):
        """Default click implementation doesn't do anything."""
        for listener in self._listeners:
            listener(event)

    def click(self):
        # Make an event to pass to click handler.
        event = GActionEvent(None, None, None)
        self._click(event)


class GCheckBox(GInteractor):
    """An onscreen check box.

    Clicking once on the check box selects it; clicking again removes the
    selection.

    Clicking on the box generates a GActionEvent.
    """

    def __init__(self, label):
        """
        Creates a GCheckBox with the specified label.

        In contrast to the GButton constructor, this constructor does not set an
        action command.

        :param str label: A label for this GCheckBox.
        """
        super().__init__()
        self.label = label
        _platform.Platform().gcheckbox_constructor(self, label)


    @property
    def selected(self):
        """Get or set whether the checkbox is selected.

        :param bool state: Whether the checkbox is selected.
        """
        return _platform.Platform().gcheckbox_is_selected(self)

    @selected.setter
    def selected(self, state):
        return _platform.Platform().gcheckbox_set_selected(self, state)

    def __str__(self):
        return 'GCheckBox("{}")'.format(self.label)


class GSlider(GInteractor):
    """An onscreen slider.

    Dragging
    """

    def __init__(self, min_value=0, max_value=100, starting_value=50):
        """Create a horizontal GSlider.

        The client can specify a minimum value, maximum value, and starting
        value.

        :param int min_value: The minimum value of this GSlider.
        :param int max_value: The maximum value of this GSlider. Should be >=
            min_value, but this is not currently enforced.
        :param int starting_value: The starting value of this GSlider. Should be
            between min_value and max_value. If not, is set to min_value.
        """
        super().__init__()
        self.min_value = min_value
        self.max_value = max_value
        # TODO(sredmond): Error check that max_value >= min_value.
        if not min_value <= starting_value <= max_value:
            starting_value = min_value
        _platform.Platform().gslider_constructor(self, min_value, max_value, starting_value)

    @property
    def value(self):
        """Get or set the current value of the GSlider.

        :param int value: The value of the GSlider.
        """
        return _platform.Platform().gslider_get_value(self)

    @value.setter
    def value(self, new_value):
        _platform.Platform().gslider_set_value(self, value)

    def __str__(self):
        return "GSlider(value={}, min={}, max={})".format(self.value, self.min_value, self.max_value)


class GTextField(GInteractor):
    """An onscreen text field for entering short text strings.

    Hitting RETURN in a text field generates a :class:`GActionEvent`.
    """

    def __init__(self, width=10):
        """Create a text field with a maximum width.

        A :class:`GActionEvent` is generated whenever the user presses the
        RETURN key while this interactor has focus.

        :param width: The maximum number of characters this field can hold.
        """
        super().__init__(self)
        self._width = width
        _platform.Platform().gtextfield_constructor(self)

    @property
    def text(self):
        """Get or set the contents of this :class:`GTextField`."""
        return _platform.Platform().gtextfield_get_text(self)

    @text.setter
    def text(self, content):
        _platform.Platform().gtextfield_set_text(self, content)

    def __str__(self):
        return "GTextField(text={}, width={})".format(self.text, self.width)


class GChooser(GInteractor):
    """A list of selectable items.

    You can construct a :class:`GChooser` with an ordered collection of items,
    or :meth:`add` and :meth:`remove` items after construction::

        size_chooser = GChooser('Small', 'Medium', 'Large')
        size_chooser.add_item('X-Large')
        size_chooser.remove_item('Medium')

    To get the selected item::

        size_chooser = GChooser('Small', 'Medium', 'Large')
        selected = size_chooser.selected_item
    """

    def __init__(self, *items):
        """Create a :class:`GChooser`, optionally with some items.

        These items are supplied as extra positional arguments::

            size_chooser = GChooser('Small', 'Medium', 'Large')

        To construct a :class:`GChooser` from a collection of elements, unpack
        the elements into the constructor::

            options = ('Small', 'Medium', 'Large')
            size_chooser = GChooser(*options)

        A :class:`GActionEvent` is generated each time the user selects an item.
        """
        super().__init__(self)
        self._items = list(items)

        _platform.Platform().gchooser_constructor(self)

    def add_item(self, item):
        """Add a new option to this :class:`GChooser`.

        :param item: The new item to add to this :class:`GChooser`.
        """
        self._items.append(item)
        _platform.Platform().gchooser_add_item(self, item)

    def remove_item(self, item):
        """Remove an option from this :class:`GChooser`.

        :param item: The item to remove from this :class:`GChooser`.
        :return: Whether the item was removed.
        """
        try:
            self._items.remove(item)
            _platform.Platform().gchooser_remove_item(self, item)
            return True
        except ValueError:
            return False

    @property
    def selected_item(self):
        """Get or set the currently selected item of this :class:`GChooser`.

        Setting the selected item to something not in this :class:`GChooser`'s
        list of items results in no change.
        """
        return _platform.Platform().gchooser_get_selected_item(self)

    @selected_item.setter
    def selected_item(self, item):
        if item not in self._items:
            return

        _platform.Platform().gchooser_set_selected_item(self, item)

    def __str__(self):
        return "GChooser(items={})".format(self._items)



#!/usr/bin/env python3 -tt
"""

"""
class Grid(list):
    def __init__(self, num_rows=0, num_cols=0, value=None):
        self.num_rows = num_rows
        self.num_cols = num_cols
        # Elements are stored in row-major order
        self.elems = [[value for __ in range(num_cols)] for _ in range(num_rows)]

    def empty(self):
        return self.num_rows == 0 or self.num_cols == 0

    def fill(self, value):
        self.elems = [[value for __ in range(self.num_cols)] for _ in range(self.num_rows)]

    def get(self, row, col):
        return self.elems[row][col]

    def set(self, row, col, value):
        self.elems[row][col] = value

    def in_bounds(self, row, col):
        pass

    def resize(self, num_rows, num_cols, retain=False):
        if not retain:
            self.__init__(num_rows, num_cols)
        else:
            old = self.elems
            self.elems = [[None for __ in range(self.num_cols)] for _ in range(self.num_rows)]
            for oldrow, newrow in zip(old, self.elems):
                for pos, elems in enumerate(zip(oldelems, newelems)):
                    oldelem, newelem = elems
                    if oldelem != newelem:
                        newrow[pos] = oldelem

    # Use this when you need to copy
    @classmethod
    def from_grid(cls, grid):
        new = cls(grid.num_rows, grid.num_cols)
        new.elems = grid.elems
        return new

    @property
    def height(self):
        return num_rows

    @property
    def width(self):
        return num_cols

    # Magic methods to make a grid look like a list of lists

    def __len__(self):
        return len(self.elems)

    # Note - setitem and delitem don't make sense in this case
    def __getitem__(self, key):
        return self.elems[key]

    def __iter__(self):
        return iter(self.elems)

    def __reversed__(self):
        return reversed(self.elems)

    def __contains__(self, item):
        return item in self.elems

    def __repr__(self):
        return repr(self.elems)




"""CAMPY provides Stanford's graphical libraries in Python.

This setup.py file is modified from:
- https://packaging.python.org/guides/distributing-packages-using-setuptools/
- https://github.com/pypa/sampleproject
"""
# Always prefer setuptools over distutils
from setuptools import setup, find_packages
from os import path
from io import open  # Default to text mode with universal newlines.

# Load the text of the README file.
HERE = path.abspath(path.dirname(__file__))
with open(path.join(HERE, 'README.md'), encoding='utf-8') as f:
    README_TEXT = f.read()

setup(
    # The name of this project, as registered on PyPI.
    # You can install this package:
    #     $ pip install campy
    # It lives on PyPI at https://pypi.org/project/campy/
    name='campy',

    # Project versions, compliant with PEP440.
    # TODO(sredmond): Single-source the project version.
    # https://packaging.python.org/guides/single-sourcing-package-version/
    version="0.0.1.dev11",

    description="Stanford's introductory libraries in Python, including the ACM graphical libraries.",
    long_description=README_TEXT,
    long_description_content_type="text/markdown",

    # The project's main homepage.
    url='https://campy.sredmond.io',

    # Author details.
    author='Sam Redmond',
    author_email='sredmond@stanford.edu',

    # Project license.
    license='MIT',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',

        # Project License (should match the license above).
        'License :: OSI Approved :: MIT License',

        'Topic :: Education',
        'Topic :: Software Development :: User Interfaces',

        # Supported Python versions.
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],

    # Publicly visible project keywords.
    keywords='Stanford teaching education intro introductory ACM graphics CS106A CS106B CS106X CS106AP Karel GObject GOval GRect GWindow',

    # Exported packages.
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),

    # Supported Python versions. Unlike the classifiers above, `pip install`
    # will check this and refuse to install the project if the version does not
    # match.
    python_requires='>=3.4, <4',

    # Run-time dependencies, installed by pip when this project is installed.
    install_requires=[],  # No dependencies. Woohoo!

    # Additional dependency groups.
    # Install these using the following syntax, for example:
    #     $ pip install -e .[dev,test]
    extras_require={
        'dev': ['check-manifest', 'pycodestyle'],
        'test': ['tox', 'pytest', 'pytest-cov', 'coverage'],
    },

    # Include data files in the package.
    # TODO(sredmond): Include spl.jar if SPL support is installed, somehow.
    # package_data={
    #     'sample': ['package_data.dat'],
    # },

    # Executable scripts (as entry points, preferred to scripts), providing
    # cross-platform support with the appropriate executable for the target
    # platform.
    # entry_points={
    #     'console_scripts': [
    #         'sample=sample:main',
    #     ],
    # },

    # Additional URLs.
    project_urls={
        'Source': 'https://github.com/sredmond/campy/',
        'Bug Reports': 'https://github.com/sredmond/campy/issues',
        'Say Thanks!': 'http://saythanks.io/to/sredmond',
    },
)



"""
This file defines the <code>GOptionPane</code> class which supports
popping up graphical dialog boxes for user input.

This class is inspired by, extremely similar to, and implemented on
the back-end by, Java's JOptionPane class, so you may wish to consult
that class's documentation in the Java API Specification for more
information.
"""
import enum as _enum

import campy.private.platform as _platform
from campy.system.error import error as error

"""Constants for showConfirmDialog types, taken from Java's JOptionPane."""
@_enum.unique
class ConfirmType(_enum.Enum):
    """The three types of confirm dialogs: Yes/No, Yes/No/Cancel, or OK/Cancel."""
    YES_NO = 0
    YES_NO_CANCEL = 1
    OK_CANCEL = 2

@_enum.unique
class ConfirmResult(_enum.Enum):
    """The various results that can be returned from some option dialogs."""
    CANCEL = -1  # For yes/no/cancel
    NO = 0  # No is 'falsey'
    YES = 1  # Yes is 'truthy'
    OK = 2  # For ok/cancel dialogs

@_enum.unique
class MessageType(_enum.Enum):
    ERROR = 0
    INFORMATION = 1
    PLAIN = -1
    QUESTION = 3
    WARNING = 2

class GOptionPane:

    class InternalResult(_enum.Enum):
        """The results that can come back from showConfirmDialog.

        These are converted into Result enum values.
        """
        CANCEL_OPTION = 2
        CLOSED_OPTION = -1
        NO_OPTION = 1
        OK_OPTION = 0
        YES_OPTION = 0

    @classmethod
    def show_confirm_dialog(cls, message, title="", confirm_type=ConfirmType.YES_NO):
        if confirm_type not in ConfirmType:
            error("GOptionPane::show_confirm_dialog: Illegal dialog type.")

        if not title:
            title = "Select an option"

        result = cls.InternalResult(_platform.Platform().goptionpane_show_confirm_dialog(message, title, confirm_type))
        print("from platform: ", result)
        if result in [cls.InternalResult.OK_OPTION, cls.InternalResult.YES_OPTION]:
            # this is weird code because JOptionPane thinks of OK and Yes as the same,
            #and differentiates based on context of whether this is an OK/Cancel or Yes/No dialog
            return ConfirmResult.OK if confirm_type == ConfirmType.OK_CANCEL else ConfirmResult.YES
        elif result == cls.InternalResult.NO_OPTION:
            return ConfirmResult.NO
        else:
            return ConfirmResult.CANCEL

    @classmethod
    def show_input_dialog(cls, message, title=""):
        return _platform.Platform().goptionpane_show_input_dialog(message, title)

    @classmethod
    def show_message_dialog(cls, message, title="", message_type=MessageType.PLAIN):
        if message_type not in MessageType:
            error("GOptionPane::showMessageDialog: Illegal dialog type.");
        if not title:
            title = "Message"
        _platform.Platform().goptionpane_show_message_dialog(message, title, message_type)

    @classmethod
    def show_option_dialog(cls, message, options, title="", initially_selected=""):
        if not title:
            title = "Select an option"
        index = _platform.Platform().goptionpane_show_option_dialog(message, title, options, initially_selected)
        if cls.InternalResult(index) != cls.InternalResult.CLOSED_OPTION and 0 <= index < len(options):
            return options[index]
        return ""

    @classmethod
    def show_text_file_dialog(cls, message, title="", rows=-1, cols=-1):
        if not title:
            title = "Text file contents"
        _platform.Platform().goptionpane_show_text_file_dialog(message, title, rows, cols)






"""An enumerated type :class:`Direction` whose elements are the four cardinal directions:
``NORTH``, ``SOUTH``, ``EAST``, and ``WEST``.

A :class:`Direction` knows what :class:`Direction` is to the left, the right, or opposite itself
using the methods :func:`left`, :func:`right`, and :func:`opposite`.

Usage::

    direction = Direction.NORTH
    direction.left()      # => Direction.WEST
    direction.right()     # => Direction.EAST
    direction.opposite()  # => Direction.SOUTH
"""
import enum

@enum.unique
class Direction(enum.Enum):
    """Represent the four compass directions."""
    NORTH = 0
    EAST = 1
    SOUTH = 2
    WEST = 3

    def left(self):
        """Return the direction to the left."""
        return self.__class__((self.value + 3) % 4)

    def right(self):
        """Return the direction to the right."""
        return self.__class__((self.value + 1) % 4)

    def opposite(self):
        """Return the opposite direction."""
        return self.__class__((self.value + 2) % 4)

    def __repr__(self):
        # Only show Direction.NAME, instead of the default '<Direction.NAME: value>`.
        return "{}.{}".format(self.__class__.__name__, self.name)

__all__ = ['Direction']



"""
File: console.py
----------------
This file redirects the <code>cin</code>, <code>cout</code>,
and <code>cerr</code> channels to use a console window.  This file
must be included in the source file that contains the <code>main</code>
method, although it may be included in other source files as well.
"""

import campy.private.platform as _platform
import enum as _enum

# Set this to True for autograding to avoid the startup of loading a graphical window.
__DONT_ENABLE_GRAPHICAL_CONSOLE = False

class Console:
    def __init__(self):
        self.console_clear_enabled = False
        self.console_echo = False
        self.console_event_on_close = False
        self.console_exit_program_on_close = False
        self.console_location_saved = False
        self.console_locked = False
        self.console_close_operation = ConsoleCloseOperation.CONSOLE_HIDE_ON_CLOSE

    def clear_console(self):
        msg = "==================== (console cleared) ===================="
        if self.console_clear_enabled:
            print(msg)
            _platform.getPlatform().jbeconsole_clear()
        else:
            print(msg)

    @property
    def console_print_exceptions(self):
        pass

    # @console_close_operation.setter
    # def console_close_operation(self, op):
    #     if self.console_locked: return
    #     self.console_close_operation = op
    #     print('TODO: jbeconsole_setCloseOperation')
    #     #_platform.getPlatform().jbeconsole_setCloseOperation(op)
    #     self.console_exit_program_on_close = op == ConsoleCloseOperation.CONSOLE_EXIT_ON_CLOSE

    # @console_echo.setter
    # def console_echo(self, echo):
    #     if self.console_locked: return
    #     self.console_echo = echo

    # @console_error_color.setter
    # def console_error_color(self, color): # TODO: annotate color w/ str
    #     if self.console_locked: return
    #     self.console_error_color = color
    #     print('TODO: jbeconsole_setConsoleColor')

    # @console_event_on_close.setter
    # def console_event_on_close(self, event_on_close):
    #     if self.console_locked: return
    #     self.console_event_on_close = event_on_close


class ConsoleCloseOperation(_enum.IntEnum):
    CONSOLE_DO_NOTHING_ON_CLOSE = 0
    CONSOLE_HIDE_ON_CLOSE = 1
    CONSOLE_DISPOSE_ON_CLOSE = 2
    CONSOLE_EXIT_ON_CLOSE = 3

# TODO: redirect stuff

_inst = Console()

